part of hakkit.hakk.builder.basic;

/**
 * Represents a literal value or a reference to a value in a hakkit memory block.
 */
class _HakkObject extends _OperatorClass
{
  final String _reference;
  final int _index;  
  final _literal;
   
  _HakkObject.literal(this._literal) : _reference = null, _index = null;
  
  _HakkObject.reference(this._reference, this._index) : _literal = null;
  
  bool get isLiteral => _literal != null;
  
  /// Operator performs an assignment to a hakk memory reference.  
  /// 
  /// It can be though of as "The result of the right-hand side goes into the 
  /// memory location on the left-hand side."
  /// 
  /// ## Examples ##
  ///      // generates code that assigns literal 5 to memory location core0.
  ///      core0 << 5; 
  ///      
  ///      // Assigns the value in core0 to core1.  This is effectively a copy operation.
  ///      core1 << core0; 
  ///      
  ///      // Assigns the result of the addition to core2.
  ///      core2 << core1 + core0; 
  ///      
  ///      // When literals are used in an expression, they must be wrapped in val().
  ///      core3 << val(10) * core2;
  ///      
  ///      // Assigns 42 to slot 0 of a memory block named 'm1' that was previously declared for the hakk.
  ///      mem('m1', 0) << 42;
  /// 
  /// param next can be a literal value or a memory reference.
  void operator <<(next){
    if (this.isLiteral){
      throw 'Cannot assign to a literal value.  Use a reference like "core0" or another memory reference.';
    }
    next = _wrapLiteral(next);
    if (isLiteral){
      throw 'Cannot using assignment operator on literal.';
    }
    
    if (next is _HakkObject){
      addLine('set: [${_reference}.${_index}, ${next}]');
    }else if (next is _ExpressionNode){
      next.compile();
      addLine('set: [${_reference}.${_index}, ${next.memLoc}]');
    }
    
    if (_globalContext.isEmpty){
      _MMU.dalloc('expression');
    }
  }
   
  static void addLine(String line){
    if (_currentCodeBlock == null){
      // implicity add the first code block
      codeEntry();
    }
    
    _currentCodeBlock.lines.add(line);
  }
  
  @override String toString(){
    if (isLiteral){
      return _literal.toString();
    }
    
    return '${_reference}.${_index}';
    
  }
}