part of hakkit.hakk.builder.basic;

/**
 * Contains blocks that can be used in constructing a hakk.  Blocks are physical components that
 * build up the "shape" of a hakk.  Each block brings with it different capabilities.
 */
class Block{
  static int _sysBlockIndex = 0;
  
  /// This is a required block for all hakks written with the hakkBASIC API.  It will 
  /// need to be declared at least once, and possibly more than once if more memory 
  /// is required to complete the compilation.
  /// 
  /// @param x is the x location of the block, relative to the core block which always exists at (0, 0).
  /// @param y is the y location of the block, relative to the core block which always exists at (0, 0).
  static void system(int x, int y){
    final sb = new _MemoryBlock('__sys${_sysBlockIndex++}__', new Point(x, y));
    _blockDeclarations.add(sb);
    _MMU.addMemoryRange(sb.name, 0, 128);
  }
  
  /// Declares an instruction block, which can contain code.
  /// 
  /// @param name is the name of the block, which is referenced later in the code sections.
  /// @param x is the x location of the block, relative to the core block which always exists at (0, 0).
  /// @param y is the y location of the block, relative to the core block which always exists at (0, 0).
  static void instructions(String name, int x, int y) => _blockDeclarations.add(new _InstructionBlock(name, new Point(x, y)));
  
  
  /// Declares a memory block, which provides extra storage space for information.
  ///
  /// @param name is the name of the block, which is referenced later in the code sections.
  /// @param x is the x location of the block, relative to the core block which always exists at (0, 0).
  /// @param y is the y location of the block, relative to the core block which always exists at (0, 0).
  static void memory(String name, int x, int y) => _blockDeclarations.add(new _MemoryBlock(name, new Point(x, y)));
  
  
  /// Declares an offensive block that, when fired, can apply heat to a cell on the playfield.  Blocks that occupy
  /// a cell with heat will begin to heat up until incapacitated.
  ///
  /// @param name is the name of the block, which is referenced later in the code sections.
  /// @param x is the x location of the block, relative to the core block which always exists at (0, 0).
  /// @param y is the y location of the block, relative to the core block which always exists at (0, 0).
  static void thermalInducer(String name, int x, int y) => _blockDeclarations.add(new _MemoryBlock(name, new Point(x, y)));
  
  
  /// Declares an offensive block that, when fired, can apply static to a cell on the playfield.  Blocks that occupy
  /// a cell with static will operate more slowly inside the hakk.it virtual machine.
  ///
  /// @param name is the name of the block, which is referenced later in the code sections.
  /// @param x is the x location of the block, relative to the core block which always exists at (0, 0).
  /// @param y is the y location of the block, relative to the core block which always exists at (0, 0).
  static void staticInducer(String name, int x, int y) => _blockDeclarations.add(new _MemoryBlock(name, new Point(x, y)));
  
  /// Declares an defensive block that, when activated, can reduce the static effect on a block.
  ///
  /// @param name is the name of the block, which is referenced later in the code sections.
  /// @param x is the x location of the block, relative to the core block which always exists at (0, 0).
  /// @param y is the y location of the block, relative to the core block which always exists at (0, 0).
  static void staticDischarger(String name, int x, int y) => _blockDeclarations.add(new _MemoryBlock(name, new Point(x, y)));
  
  /// Declares an defensive block that, when activated, can reduce the heat level of a block.
  ///
  /// @param name is the name of the block, which is referenced later in the code sections.
  /// @param x is the x location of the block, relative to the core block which always exists at (0, 0).
  /// @param y is the y location of the block, relative to the core block which always exists at (0, 0).
  static void heatSink(String name, int x, int y) => _blockDeclarations.add(new _MemoryBlock(name, new Point(x, y)));
  
  /// Declares a defensive block, than can reboot a block that has become disabled (usually by overheating).
  ///
  /// @param name is the name of the block, which is referenced later in the code sections.
  /// @param x is the x location of the block, relative to the core block which always exists at (0, 0).
  /// @param y is the y location of the block, relative to the core block which always exists at (0, 0).
  static void rebooter(String name, int x, int y) => _blockDeclarations.add(new _MemoryBlock(name, new Point(x, y)));
}

class _Rebooter extends _BlockDeclaration
{
  _Rebooter(String name, Point loc) : super(name, loc);
  
  @override toString() => '  - rebt: {name: "${name}", loc: [${this.position.x}, ${this.position.y}]}$_CR';
}

class _StaticDischarger extends _BlockDeclaration
{
  _StaticDischarger(String name, Point loc) : super(name, loc);
  
  @override toString() => '  - sdis: {name: "${name}", loc: [${this.position.x}, ${this.position.y}]}$_CR';
}

class _HeatSink extends _BlockDeclaration
{
  _HeatSink(String name, Point loc) : super(name, loc);
  
  @override toString() => '  - hsnk: {name: "${name}", loc: [${this.position.x}, ${this.position.y}]}$_CR';
}

class _ThermalInducer extends _BlockDeclaration
{
  _ThermalInducer(String name, Point loc) : super(name, loc);
  
  @override toString() => '  - ti: {name: "${name}", loc: [${this.position.x}, ${this.position.y}]}$_CR';
}

class _StaticInducer extends _BlockDeclaration
{
  _StaticInducer(String name, Point loc) : super(name, loc);
  
  @override toString() => '  - si: {name: "${name}", loc: [${this.position.x}, ${this.position.y}]}$_CR';
}

class _MemoryBlock extends _BlockDeclaration
{
  _MemoryBlock(String name, Point loc) : super(name, loc);
  
  @override toString() => '  - mem: {name: "${name}", loc: [${this.position.x}, ${this.position.y}]}$_CR';
}

class _InstructionBlock extends _BlockDeclaration
{
  _InstructionBlock(String name, Point loc) : super(name, loc);
  
  @override toString() => '  - instr: {name: "${name}", loc: [${this.position.x}, ${this.position.y}]}$_CR';
}

class _Core extends _BlockDeclaration
{
  _Core() : super('core', new Point(0, 0));
  
  @override toString() => '  - core:$_CR';
}

abstract class _BlockDeclaration
{
  final String name;
  final Point<int> position;

  _BlockDeclaration(this.name, this.position);
}