part of hakkit.hakk.builder.basic;

String _subroutineContext;


/// Orients the hakk to the given [direction].
/// 
/// This is an expensive operation.
/// 
/// @param direction must be one of the for allowed directions [NORTH], [SOUTH], 
/// [EAST], [WEST].
/// @returns a memory reference containing a boolean value that indicates if the
/// hakk was successfully turned to the desired direction.  Will also return false
/// if the hakk was already facing the given direction.
_HakkObject faceDirection(_HakkObject direction){
  _HakkObject d = _MMU.malloc('face_direction');
  _HakkObject result = _MMU.malloc('face_direction');
  result << false;  
  d << facing();
  final nl = _nextLabelID;
  
  IF (d == NORTH);
    IF (direction == SOUTH);
      result << turnLeft();
      result << turnLeft();
    ENDIF();
    IF (direction == EAST);
      result << turnRight();
    ENDIF();
    IF (direction == WEST);
      result << turnLeft();
    ENDIF();
    GOTO('__end_turn_to_${nl}__');
  ENDIF();
  
  IF (d == EAST);
    IF (direction == NORTH);
      result << turnLeft();
    ENDIF();
    IF (direction == SOUTH);
      result << turnRight();
    ENDIF();
    IF (direction == WEST);
      result << turnLeft();
      result << turnLeft();
    ENDIF();
    GOTO('__end_turn_to_${nl}__');
  ENDIF();
  
  IF (d == SOUTH);
    IF (direction == NORTH);
      result << turnLeft();
      result << turnLeft();
    ENDIF();
    IF (direction == EAST);
      result << turnLeft();
    ENDIF();
    IF (direction == WEST);
      result << turnRight();
    ENDIF();
    GOTO('__end_turn_to_${nl}__');
  ENDIF();
  
  IF (d == WEST);
    IF (direction == NORTH);
      result << turnRight();
    ENDIF();
    IF (direction == EAST);
      result << turnRight();
      result << turnRight();
    ENDIF();
    IF (direction == SOUTH);
      result << turnLeft();
    ENDIF();
  ENDIF();
  
  label('__end_turn_to_${nl}__');
  
  _MMU.dalloc('face_direction');
  return result;
}


/// Determines the direction the hakk is facing. See [NORTH], [SOUTH], [EAST], [WEST]
/// for an explaination of the values.
/// 
/// This is an expensive operation.
/// 
/// # Example
///      // stores the direction the hakk is facing into core0
///      core0 << facing();
/// 
/// # Related
/// * [NORTH]
/// * [SOUTH]
/// * [EAST]
/// * [WEST]
/// 
/// @returns a memory reference that contains a value indicating which
/// direction the hakk is facing.
_HakkObject facing(){
  _HakkObject.addLine('hvec:');
  _HakkObject x = _MMU.malloc('facing');
  _HakkObject y = _MMU.malloc('facing');
  _HakkObject result = _MMU.malloc('facing');
  
  x << pop();
  y << pop();
  IF((equals(x, val(0))) & (y == val(-1)));
    result << NORTH;
  ENDIF();
  
  IF((equals(x, val(0))) & (y == val(1)));
    result << SOUTH;
  ENDIF();
  
  IF((equals(x, val(1))) & (y == val(0)));
    result << EAST;
  ENDIF();
  
  IF((equals(x, val(-1))) & (y == val(0)));
    result << WEST;
  ENDIF();
  
  _MMU.dalloc('facing');
  return result;
}

/// Calculates the vector of the hakk and pushes the y and x components onto the 
/// stack, respectively.  
/// 
/// This is a lower level (but much faster) version of [facing].
/// 
/// # How to interpret vector results
/// Once you have the vector values, you can determine which direction the hakk
/// is facing like so:
/// 
/// * x = 0, y = -1: north
/// * x = 0, y = 1: south
/// * x = -1, y = 0: west
/// * x = 1, y = 0: east
/// 
/// # Example
///      vector();
///      // x
///      core0 << pop();
///      
///      // y
///      core1 << pop();
void vector(){
  _HakkObject.addLine('hvec:');
}

/// Calculates the x,y position of the hakk and pushes the y and x components onto 
/// the stack, respectively.  The coordinates represent the position of the 
/// core block of the hakk.
/// 
/// # Example
///      location();
///      
///      // x
///      core0 << pop();
///      
///      // y
///      core1 << pop();
void location(){
  _HakkObject.addLine('hloc:');
}

/// Calculates a normalized vector to the given [x],[y] location, from the location
/// of the core block of the hakk.  The vector results are pushed onto the stack in
/// y,x order.
/// 
/// # Example
///      vectorTo(5, 5);
///      
///      core0 << pop(); // x component of vector
///      core1 << pop(); // y component of vector
void vectorTo(x, y){
  x = _wrapLiteral(x);
  y = _wrapLiteral(y);
  
  if (x is _ExpressionNode){
    x.compile();
  }
  
  if (y is _ExpressionNode){
    y.compile();
  }
  
  _HakkObject.addLine('tvec: [${x}, ${y}]');
}

/// Converts an x,y vector to a cardinal direction value [NORTH], [SOUTH], [EAST],
/// [WEST].  If the x/y vector is invalid, then the value returned will be -1.
/// 
/// # Valid vectors
/// * x = 0, y = -1: north
/// * x = 0, y = 1: south
/// * x = -1, y = 0: west
/// * x = 1, y = 0: east
/// 
/// # Example
/// In the following example, the vectorToDirection() is used to output the four
/// cardinal directions to console.  Also demonstrated is a subroutine, using 
/// [CALL], [SUB], and [RETURN], along with parameter passing.
/// 
///     hakk('example');
///     Block.system(0, 1);
///
///     CALL('dovec', [0, -1]);
///     CALL('dovec', [0, 1]);
///     CALL('dovec', [-1, 0]);
///     CALL('dovec', [1, 0]);
///  
///     SUB('dovec', [core2, core3]);
///       core4 << vectorToDirection(core2, core3);
///       IF (core4 == NORTH);
///         output('north');
///       ENDIF();
///       IF (core4 == SOUTH);
///         output('south');
///       ENDIF();
///       IF (core4 == EAST);
///         output('east');
///       ENDIF();
///       IF (core4 == WEST);
///         output('west');
///       ENDIF();
///       IF (core4 == -1);
///         output('invalid vector!');
///       ENDIF();
///     RETURN();
/// 
/// # Related
/// * [vectorTo]
/// * [facing]
/// * [faceDirection]
/// * [vector]
/// 
/// @param x is the x value of the vector to convert.
/// @param y is the y value of the vector to convert.
_HakkObject vectorToDirection(x, y){
  x = _wrapLiteral(x);
  y = _wrapLiteral(y);
  _HakkObject result = _MMU.malloc('vector_to_direction');
  result << -1;
  
  if (x is _ExpressionNode){
    x.compile();
  }
  
  if (y is _ExpressionNode){
    y.compile();
  }
  
  IF((equals(x, val(0))) & (y == val(-1)));
    result << NORTH;
  ENDIF();
  
  IF((equals(x, val(0))) & (y == val(1)));
    result << SOUTH;
  ENDIF();
  
  IF((equals(x, val(1))) & (y == val(0)));
    result << EAST;
  ENDIF();
  
  IF((equals(x, val(-1))) & (y == val(0)));
    result << WEST;
  ENDIF();
  
  _MMU.dalloc('vector_to_direction');
  return result;
}

/// Locates the nearest enemy hakk on the playfield, and pushes the y,x coordinates
/// of the enemy core block onto the stack, respectively.
/// 
/// 
void locateEnemy(){
  _HakkObject.addLine('find:');
}

_HakkObject faceEnemy(){
  _HakkObject result = _MMU.malloc('face_enemy');
  _HakkObject x = _MMU.malloc('face_enemy');
  _HakkObject y = _MMU.malloc('face_enemy');
  
  locateEnemy();
  x << pop();
  y << pop();
  output('enemy location');
  vectorTo(x, y);
  x << pop();
  y << pop();
  output('vector to enemy:');
  result << faceDirection(vectorToDirection(x, y));
  
  _MMU.dalloc('face_enemy');
  return result;
}

_ExpressionNode moveForward(){
  final node = new _ExpressionNode(val(0), val(0), 'movef');
  _expressionQueue.add(node);
  return node;
}

_ExpressionNode moveBack(){
  final node = new _ExpressionNode(val(0), val(0), 'moveb');
  _expressionQueue.add(node);
  return node;
}

_ExpressionNode turnLeft(){
  final node = new _ExpressionNode(val(0), val(0), 'turnl');
  _expressionQueue.add(node);
  return node;
}

_ExpressionNode turnRight(){
  final node = new _ExpressionNode(val(0), val(0), 'turnr');
  _expressionQueue.add(node);
  return node;
}


/// Causes execution to halt for [expr] number of ticks in the hakk.it virtual
/// machine.
/// 
/// # Example
///     hakk('wait_example');
///     Block.system(0, 1);
///     
///     output('hi');
///     
///     // wait 60 ticks (about 1 second)
///     wait(60);
///     
///     output('there');
///     
///     core0 << 5;
///     core1 << 25;
///     
///     // wait supports expressions...
///     wait(core0 + core1);
///     output('finished.');
void wait(expr){
  expr = _wrapLiteral(expr);
  if (expr is _ExpressionNode){
    expr.compile();
    _HakkObject.addLine('wait: ${expr.memLoc}');
  }else{
    _HakkObject.addLine('wait: ${expr}'); 
  }
}

/// Outputs [expr] to the STDOUT (usually the console).
/// 
/// This instruction doesn't really help in the arena, but is sometimes useful
/// for testing and demos.  Use sparingly in hakks intended for arena combat, as
/// output will waste precious ticks.
/// 
/// # Example
///     hakk('output_example');
///     Block.system(0, 1);
///     
///     // output a string
///     output('Hello World!');
///     
///     // output a numeric literal
///     output(42);
///     
///     // output a boolean
///     output(true);
///     
///     // output an expression
///     core0 << 5;
///     core1 << 15;
///     output(core0 + core1);
///     // "15"
void output(expr){
  expr = _wrapLiteral(expr);
  if (expr is _HakkObject){
    if (expr.isLiteral){
      _HakkObject.addLine('out: "${expr}"');
    }else{
      _HakkObject.addLine('out: ${expr}');
    }
  }else if (expr is _ExpressionNode){
    expr.compile();
    _HakkObject.addLine('out: ${expr.memLoc}');
  }else{
    throw 'Invalid type passed to output.  This is probably a bug on our end. Sorry!';
  }
  if (_globalContext.isEmpty){
    _MMU.dalloc('expression');
  }
}

/// Increments a value by 1;
/// 
/// Equivalent to (using [core0] as an example):
/// 
///      core0 << core0 + 1;
///      
/// # Related
/// * [dec]
/// 
/// @param expr is a memory location with a value to be incremented.
void inc(expr){
  expr = _wrapLiteral(expr);
  if (expr.isLiteral){
    throw "The inc function only works on values in a memory reference.  Supply a memory reference instead.";
  }
  expr <<  expr + 1;
  if (_globalContext.isEmpty){
    _MMU.dalloc('expression');
  }
}

/// Decrements a value by 1;
/// 
/// Equivalent to (using [core0] as an example):
/// 
///      core0 << core0 - 1;
///      
/// # Related
/// * [inc]
/// 
/// @param expr is a memory location with a value to be incremented.
void dec(expr){
  expr = _wrapLiteral(expr);
  if (expr.isLiteral){
    throw "The inc function only works on values in a memory reference.  Supply a memory reference instead.";
  }
  expr << expr - 1;
  if (_globalContext.isEmpty){
    _MMU.dalloc('expression');
  }
}

/// Inlines hakkML into the source code.
/// 
/// ## Use With Caution
/// This is an advanced feature, and should not be used unless you have a really good
/// understanding of how the basic API builder works, along with a strong working
/// knowledge of hakkML.  The builder makes certain assumptions about the structure 
/// of code it generates, and inlining hakkML may break this.
/// 
/// ## Example
///     hakk('inline_example');
///     Block.system(0, 1);
///     
///     core0 << 42;
///     // outputs "42";
///     output(core0);
///     
///     //change core0 with inline
///     inline('set: [core.0, 24]');
///     
///     // outputs "24";
///     output(core0);
///     
///     // print the compiled hakkML to the console.
///     print(build());
///     
/// @param instructions If a string, is assumed to be a single instruction.  If a 
/// list of Strings, then is assumed to be a list of instructions.
void inline(instructions){
  if (instructions is String){
    _HakkObject.addLine(instructions);
  }else if (instructions is List){
    instructions.forEach((i) => _HakkObject.addLine(i));
  }
}

/// Generates a random integer between 0 (inclusive) and [expr] (exclusive)
/// 
/// # Example
///     hakk('random_example');
///     Block.system(0, 1);
///     
///     core0 << random(10);
///     output('Will output an integer between 0 - 9');
///     output(core0);
/// 
/// @param expr is a literal number or memory location of a number that represents
/// the upper bound of the random result.
_ExpressionNode random(expr){
  expr = _wrapLiteral(expr);
  final node = new _ExpressionNode(expr, expr, 'rnd');
  _expressionQueue.add(node);
  return node;
}

/// Pushes a value from [expr] onto the stack.
/// 
/// # The Stack
/// Each hakk comes with it's own value stack, allowing coders to utilize
/// techniques to pass information from one part of a hakk to another, without the
/// receiver having knowledge of "where" the information is.  Rather, the receiver
/// only needs to know how much information to [pop] off the stack.  Even the "amount"
/// can be added to the stack, provided a dynamic way to pass information around.
/// 
/// # The importance of balance
/// It's important that values pushed onto the stack are always [pop]ped from the
/// stack in balance (if you push 3 values, you should always pop 3 values elsewhere).
/// Failure to do this will likely result in stack overflow or stack underflow.
/// 
/// # Example
///     hakk('stack_example');
///     Block.system(0, 1);
///     
///     // push 3 values onto the stack, the last value being the "amount" of
///     // other values pushed.
///     core0 << 42;
///     // core0 can safely be used again because it's value is now pushed onto the
///     // stack.
///     push(core0);
///     push(-42);
///     push(2);
///     
///     // ..somewhere else in code
///     
///     core0 << pop();
///     output('how many items to retrieve:');
///     output(core0);
///     core1 << 0;
///
///     output('');
///     output('Items:');
///     WHILE(core1 <= core0);
///       core3 << pop();
///       output(core3);
///       inc(core1);
///     ENDWHILE();
///     
///     // outputs:
///     // -42
///     // 42
///  
/// # Related
/// * [pop]
void push(expr){
  expr = _wrapLiteral(expr);
  if(expr is _ExpressionNode){
    expr.compile();
    _HakkObject.addLine('push: ${expr.memLoc}');
  }else{
    _HakkObject.addLine('push: ${expr}');
  }
}

/// Pops a value off the stack.
/// 
/// See [push] for documentation of how push/pop work together.
///   
/// # Related
/// * [push]
/// 
/// @returns the value popped off the stack.
_ExpressionNode pop(){
  final node = new _ExpressionNode(val(0), val(0), 'pop');
  _expressionQueue.add(node);
  return node;
}

/// Returns the cosine of a numeric expression.
/// 
/// # Related
/// * [sine]
/// 
/// @param expr a numeric expression.
/// @returns the cosine of a numeric [expr] expression.
_ExpressionNode cosine(expr){
  expr = _wrapLiteral(expr);
  final node = new _ExpressionNode(expr, expr, 'cos');
  _expressionQueue.add(node);
  return node;
}

/// Returns the sine of a numeric expression.
/// 
/// # Related
/// * [cosine]
/// 
/// @param expr a numeric expression.
/// @returns the sine of a numeric [expr] expression.
_ExpressionNode sine(expr){
  expr = _wrapLiteral(expr);
  final node = new _ExpressionNode(expr, expr, 'sin');
  _expressionQueue.add(node);
  return node;
}

/// Calculates the result of a numeric [term] taken to a given [power].
/// 
/// Both parameters can themselves be expressions.
/// 
/// @param term is the numeric value to be operated on.
/// @param power is the exponent to apply to the term.
/// 
/// # Example
///      core0 << power(2, 4);
///      core0 << power(core0, 2);
_ExpressionNode power(term, power){
  term = _wrapLiteral(term);
  power = _wrapLiteral(power);

  final node = new _ExpressionNode(term, power, 'pow');
  _expressionQueue.add(node);
  return node;
}

/// Sets a memory [reference] to the result of a given [expr] expression.
/// 
/// Equivalent to the operator '<<' that is more commonly used.
/// 
/// # Example
///      set(core0, 42);
///      // equivalent to:
///      core0 << 42;
void set(_HakkObject reference, expr){
  expr = _wrapLiteral(expr);
  reference << expr;
}

/// Determines equivalancy between two expressions.
/// 
/// Equivalent to:
/// 
///      {expr1} == {expr2}
/// 
/// @returns true if the expressions are equal, false otherwise.
_ExpressionNode equals(expr1, expr2){
  expr1 = _wrapLiteral(expr1);
  expr2 = _wrapLiteral(expr2);
  final node = new _ExpressionNode(expr1, expr2, 'je');
  _expressionQueue.add(node);
  return node;
}

/// Represents value 0, which indicates the direction the hakk is facing
/// 
/// NORTH (0) means the hakk is facing up (toward the top) of the arena.
/// 
/// # Related
/// * [EAST]
/// * [SOUTH]
/// * [WEST]
final NORTH = val(0);
/// Represents value 0, which indicates the direction the hakk is facing
/// 
/// EAST (1) means the hakk is facing right (toward the right side of the arena).
/// 
/// # Related
/// * [NORTH]
/// * [SOUTH]
/// * [WEST]
final EAST = val(1);
/// Represents value 0, which indicates the direction the hakk is facing
/// 
/// SOUTH (2) means the hakk is facing down (toward the bottom of the arena).
/// 
/// # Related
/// * [EAST]
/// * [NORTH]
/// * [WEST]
final SOUTH = val(2);
/// Represents value 0, which indicates the direction the hakk is facing
/// 
/// WEST (3) means the hakk is facing left (toward the left side of the arena).
/// 
/// # Related
/// * [EAST]
/// * [SOUTH]
/// * [WEST]
final WEST = val(3);

/// Represents memory address 0 in the core block of a hakk.
final core0 = new _HakkObject.reference('core', 0);
/// Represents memory address 1 in the core block of a hakk.
final core1 = new _HakkObject.reference('core', 1);
/// Represents memory address 2 in the core block of a hakk.
final core2 = new _HakkObject.reference('core', 2);
/// Represents memory address 3 in the core block of a hakk.
final core3 = new _HakkObject.reference('core', 3);
/// Represents memory address 4 in the core block of a hakk.
final core4 = new _HakkObject.reference('core', 4);
/// Represents memory address 5 in the core block of a hakk.
final core5 = new _HakkObject.reference('core', 5);
/// Represents memory address 6 in the core block of a hakk.
final core6 = new _HakkObject.reference('core', 6);
/// Represents memory address 7 in the core block of a hakk.
final core7 = new _HakkObject.reference('core', 7);
/// Represents memory address 8 in the core block of a hakk.
final core8 = new _HakkObject.reference('core', 8);
/// Represents memory address 9 in the core block of a hakk.
final core9 = new _HakkObject.reference('core', 9);
/// Represents memory address 10 in the core block of a hakk.
final core10 = new _HakkObject.reference('core', 10);
/// Represents memory address 11 in the core block of a hakk.
final core11 = new _HakkObject.reference('core', 11);
/// Represents memory address 12 in the core block of a hakk.
final core12 = new _HakkObject.reference('core', 12);
/// Represents memory address 13 in the core block of a hakk.
final core13 = new _HakkObject.reference('core', 13);
/// Represents memory address 14 in the core block of a hakk.
final core14 = new _HakkObject.reference('core', 14);
/// Represents memory address 15 in the core block of a hakk.
final core15 = new _HakkObject.reference('core', 15);