part of hakkit.hakk.builder.lang;

/**
 * Grammar definition for HakkBASIC.
 * 
 * Big thanks to Petitparser examples!
 */
class HakkBasicGrammar extends CompositeParser2
{
 
  @override 
  void initialize(){
    _whitespace();
    _lexemes();
    _reserved();
    _declarations();
    _expressions();
  }
  
  /// Defines whitespace, documentation and comments.
  void _whitespace() {
    def('whitespace', whitespace()
      .or(ref('singe-line comment'))
      .or(ref('multi-line comment')));
    def('newline', Token.newlineParser());

    def('singe-line comment', string('//')
      .seq(ref('newline').neg().star())
      .seq(ref('newline').optional()));
    def('multi-line comment', string('/*')
      .seq(ref('multi-line comment')
          .or(string('*/').neg())
          .star())
      .seq(string('*/')));
  }
  
  /// Helper to define a token parser that consumes whitespace.
  Parser _t(input) {
    var parser = input is Parser 
        ? input 
        : input.length == 1 
            ? char(input) 
            : string(input);
    return parser.token().trim(ref('whitespace'));
  }
  
  // prims
  Parser DIGIT, LETTER, EXPONENT, NUMBER, ESCAPE_SEQUENCE, INTEGER;
  void _lexemes() {
    DIGIT = digit();
    LETTER = letter();

    EXPONENT = (char('e') | char('E')) & (char('+') | char('-')).optional() & DIGIT.plus();

    INTEGER =  _t('-').optional() & DIGIT.plus();
    
    NUMBER = DIGIT.plus() & (char('.') & DIGIT.plus()).optional() & EXPONENT.optional()
           | char('.') & DIGIT.plus() & EXPONENT.optional()
           ;

    ESCAPE_SEQUENCE =
        string('\\n')
      | string('\\r')
      | string('\\f')
      | string('\\b')
      | string('\\t')
      | string('\\v');

  }
  
  // reserved
  Parser ELSE, TRUE, FALSE, FOR, IF, RETURN, WHILE, EXIT, BLOCKS, CALL, GOTO, BREAK;
  void _reserved() {
    ELSE = _t('else');
    FALSE = _t('false');
    FOR = _t('for');
    IF = _t('if');
    RETURN = _t('return');
    TRUE = _t('true');
    WHILE = _t('while');
    EXIT = _t('exit');
    CALL = _t('call');
    GOTO = _t('goto');
    BREAK = _t('break');
    BLOCKS = _t('blocks');
    
    // built-in identifiers
  } 
  
  Parser binaryOperator, parameterList, operator, parameters, parameter, statement, statements;
  Parser block, functionSignature, blockDeclaration, blockStatements, declaration;
  Parser blockStatementSignature, body;
  void _declarations() {
     
      parameter = identifier | literal;
      parameters = parameter.star().separatedBy(_t(','));

//      statement = identifier & _token('=') & expression.plus();
      
      parameterList =
        _t('(') & _t(')')
        | _t('(') & parameters & _t(')')
      ;

//      statements = statement.star().separatedBy(_token(';'));
            
      functionSignature = label & parameterList;
      
      block = _t('{') & statements & _t('}');

      binaryOperator =
        multiplicativeOperator
      | additiveOperator
      | relationalOperator
      | _t('==')
      ;

      operator =
        _t('~')
      | binaryOperator
      | _t('[') & _t(']')
      | _t('[') & _t(']') & _t('=')
      ;
            
      blockDeclaration = BLOCKS & _t('{') & statements & _t('}');

      declaration =
        blockDeclaration
        | functionSignature
      ;
      
      body = blockDeclaration.optional() & (functionSignature & block).plus();
    }

  // expressions
  Parser identifier, expression, primary, unaryOperator, prefixOperator, incrementOperator;
  Parser literal, booleanLiteral, numericLiteral, assignmentOperator, compoundAssignmentOperator;
  Parser NEWLINE, expressionList, stringLiteral, stringContentDQ, stringContentSQ;
  Parser argumentList, arguments, multiplicativeOperator, unaryExpression, postfixExpression;
  Parser multiplicativeExpression, additiveOperator, additiveExpression, relationalOperator;
  Parser label, relationalExpression, memoryReference, equalityOperator;
  Parser equalityExpression, logicalAndExpression, logicalOrExpression, logicalExpression;
  Parser functionExpression;
  void _expressions() {

      memoryReference = LETTER.plus() & _t('.') & DIGIT.plus();
      
      identifier = LETTER.plus() & (DIGIT.star() | LETTER.star()).optional();

      incrementOperator = _t('++') | _t('--');
      
      postfixExpression = memoryReference & incrementOperator;     

      prefixOperator = _t('-') | _t('!');
      
      unaryExpression =
        prefixOperator & unaryExpression
      | postfixExpression
      | prefixOperator
      | incrementOperator & memoryReference
      ;

      multiplicativeOperator = _t('*') | _t('/');
      multiplicativeExpression =
        unaryExpression & (multiplicativeOperator & unaryExpression).star()
      ;

      additiveOperator = _t('+') | _t('-');

      additiveExpression =
        multiplicativeExpression & (additiveOperator & multiplicativeExpression).star()
      ;

      relationalOperator = _t('<') | _t('>') | _t('<=') | _t('>=');
      relationalExpression = expression & relationalOperator;

      equalityOperator = _t('==') | _t('!=');
      equalityExpression =
        relationalExpression & (equalityOperator & relationalExpression).optional()
      ;

      logicalAndExpression = expression & (_t('&&') & expression).star();
      logicalOrExpression = logicalAndExpression & (_t('||') & logicalAndExpression).star();

      logicalExpression = memoryReference | booleanLiteral | equalityExpression | logicalAndExpression | logicalOrExpression;
      
      compoundAssignmentOperator =
        _t('*=')
      | _t('/=')
      | _t('+=')
      | _t('-=');
      
      assignmentOperator = _t('=') | compoundAssignmentOperator;

      argumentList = expressionList.separatedBy(_t(','));
      
      arguments = _t('(') & argumentList.optional() & _t(')');

      NEWLINE = Token.newlineParser();
      
      stringContentDQ = (char('\\') | char('"') | char('\$') | NEWLINE).neg() |
          char('\\') & NEWLINE.neg();
      
      stringContentSQ = (char('\\') | char("'") | char('\$') | NEWLINE).neg() |
          char('\\') & NEWLINE.neg();

      stringLiteral =
            char('"') & stringContentDQ.star() & char('"')
          | char("'") & stringContentSQ.star() & char("'")
          ;

      functionExpression = parameterList & block;
      
      numericLiteral = _t(NUMBER);

      booleanLiteral = TRUE | FALSE;
      
      literal =
      booleanLiteral
      | numericLiteral
      | stringLiteral
      ;
          
      expression =
        additiveExpression
        | multiplicativeExpression
        | logicalExpression
      ;

      expressionList =  expression.separatedBy(_t(','));

      primary =
        functionExpression
        | literal
        | identifier
        | _t('(') & expression & _t(')')
      ;

    }
  
  Parser breakStatement, returnStatement, whileStatement, forInitializerStatement, forLoopParts;
  Parser forStatement, ifStatement, localFunctionDeclaration, nonLabelledStatement;
  Parser assignStatement;
  void _statements() {
    breakStatement = BREAK & identifier.optional() & _t(';');

    returnStatement = RETURN & expression.optional() & _t(';');

    whileStatement = WHILE & _t('(') & expression & _t(')') & statement;

    assignStatement = memoryReference & _t('=') & expression;
    
    forInitializerStatement =
      assignStatement & _t(';')
    | expression.optional() & _t(';')
    ;

    forLoopParts =
      forInitializerStatement & expression.optional() & _t(';') & expressionList.optional()
    ;

    forStatement = FOR & _t('(') & forLoopParts & _t(')') & statement;
    ifStatement = IF & _t('(') & expression & _t(')') & statement & (ELSE & statement).optional();
    localFunctionDeclaration = functionSignature & block;

    nonLabelledStatement =
          block
          | assignStatement
          | forStatement
          | whileStatement
          | ifStatement
          | breakStatement
          | returnStatement
          | localFunctionDeclaration;


    statement = label.star() & nonLabelledStatement;
    statements = statement.star();

  }
}